from simple_framework.utilities.checkpoint_saver import Checkpoint_saver
from simple_framework.utilities.metrics import AverageMeter
from simple_framework.utilities.schedulers import get_flat_cosine_schedule



import os
import torch
import logging
import sys

from typing import Dict, List

import pandas as pd

from simple_framework.trainer.BaseTrainerClass import SimpleFrameworkWrapper

from torch.utils.tensorboard import SummaryWriter
from pathlib import Path
from tqdm import tqdm

from simple_framework.backends.simple_backend import SimpleBackend
from simple_framework.backends.horovod_backend import HorovodBackend


"""
=================================CREATING TRAINER CLASS=================================
"""


class Trainer:
    def __init__(self, model: SimpleFrameworkWrapper, cfg: Dict, gpus=None):

        if isinstance(gpus, int):
            gpus_to_use = str(gpus)
        elif isinstance(gpus, List):
            gpus_to_use = " ".join([str(elem) + "," for elem in gpus])
        else:
            gpus_to_use = ""

        os.environ["CUDA_VISIBLE_DEVICES"] = gpus_to_use

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.model = model.to(device)
        self.settings = {
            "batch_size": None,
            "epochs": 1,
            "steps_per_epoch": None,
            "validation_steps": None,
            "validation_batch_size": None,
            "validation_freq": 1,
            "checkpoint_every_n_steps": None,
            "shuffle": True,
            "description": cfg["description"] if cfg["description"] is not None else "model",
            "save_path": cfg["save_path"],
            "experiment_path": cfg["experiment_path"],
            "backbone": cfg["backbone"] if cfg["backbone"] is not None else "Simple",
            "validation_metric": "loss",
            "use_fp16": False,
        }

        self.__select_backbone()

    def __select_backbone(self):
        if self.settings["backbone"] == "Simple":
            self.processing_backend = SimpleBackend(self.model)
        elif self.settings["backbone"] == "Horovod":
            print("backbone set to horovod")
            self.processing_backend = HorovodBackend(self.model)

    def fit(
        self,
        train_dataset=None,
        batch_size=None,
        epochs=1,
        validation_dataset=None,
        steps_per_epoch=None,
        validation_steps=None,
        validation_batch_size=None,
        validation_freq=1,
        checkpoint_every_n_steps=None,
        validation_metric="loss",
        shuffle=True,
        num_workers=0,
    ):
        """
        Method used for training model with following parameters:

        Parameters:
            :param torch.utils.data.Dataset train_dataset: Dataset for model.
            :param int batch_size: batch size that should be used during training/validation
            :param int epochs: Number of epochs to train
            :param torch.utils.data.Dataset validation_dataset: Dataset for validation data
            :param int steps_per_epoch: training steps that should be performed each epoch.
                If not specified whole training set would be used
            :param int validation_steps: validation steps that should be  performed each validation phase.
                If not specified, whole validation set would be used
            :param int validation_batch_size: Batch size for validation step.
                If not set, training batch size would be used
            :param int validation_freq:After how many epochs validation should be performed
            :param int checkpoint_every_n_steps: should be set if we want to save model in given timesteps interval
            :param str validation_metric: metric that should be checkd after validation to see if result is better
            :param bool shuffle: should dataset be shuffled during training
            :param int num_workers: number of workers to use during training

        Returns:
            None
        """

        assert batch_size is not None
        assert train_dataset is not None

        self.settings["batch_size"] = batch_size
        self.settings["epochs"] = epochs
        self.settings["shuffle"] = shuffle
        self.settings["validation_metric"] = validation_metric
        self.settings["num_workers"] = num_workers

        self.settings["steps_per_epoch"] = steps_per_epoch

        self.settings["validation_steps"] = validation_steps
        self.settings["validation_freq"] = validation_freq

        if validation_batch_size is None:
            self.settings["validation_batch_size"] = batch_size
        else:
            self.settings["validation_batch_size"] = validation_batch_size

        self.settings["checkpoint_every_n_steps"] = checkpoint_every_n_steps

        logging.info("Training settings:")
        for _, (key, val) in enumerate(self.settings.items()):
            logging.info(f"{key} : {val}")

        """
        ========================== TRAINING VALIDATION PHASE ==========================
        """

        self.processing_backend.setup(self.settings)
        self.processing_backend.train_phase(train_dataset, validation_dataset)



import torch
from pathlib import Path

import os
import logging


class Checkpoint_saver:
    def __init__(self, checkpoints_dir, description):
        self.best_train_metric = 10000000
        self.best_validation_metric = 10000000
        self.current_epoch = 0
        self.descritpion = description
        self.checkpoints_dir = checkpoints_dir

    def set_epoch(self, epoch):
        self.current_epoch = epoch

    def should_save_best(self, metric_value, model, checkpoint_name="best_model_checkpoint.pth", task="train"):

        metric_val = self.best_train_metric if task == "train" else self.best_validation_metric

        if metric_val > metric_value:
            if task == "train":
                self.best_train_metric = metric_value
            else:
                self.best_validation_metric = metric_value

            self.save_checkpoint(metric_value, model, checkpoint_name, task)

    def save_checkpoint(self, metric_value, model, checkpoint_name, task="train", checkpoint_path=None):

        if checkpoint_path is None:
            checkpoint_path = self.checkpoints_dir

        Path(checkpoint_path).mkdir(parents=True, exist_ok=True)
        logging.info(f"saving model for {task} with metric {metric_value} as {checkpoint_name}")
        torch.save(model.state_dict(), os.path.join(checkpoint_path, checkpoint_name))

    def save(self, checkpoint_name, model, optimizer, scheduler=None, amp=None):

        path = os.path.join(self.checkpoints_dir, "last_checkpoint.bin")
        logging.info(f"saving checkpoint to {path}")
        model.model.eval()

        save_dict = {
            "model_state_dict": model.model.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
            "epoch": self.current_epoch,
            "best_loss": self.best_validation_metric,
        }

        if scheduler is not None:
            save_dict["scheduler_state_dict"] = scheduler.state_dict()

        if amp is not None:
            save_dict["amp"] = amp.state_dict()

        torch.save(save_dict, path)

    def load(self, path, model, optimizer, scheduler=None, amp=None):
        logging.info(f"loading checkpoint from {path}")
        checkpoint = torch.load(path)
        model.model.load_state_dict(checkpoint["model_state_dict"])
        optimizer.load_state_dict(checkpoint["optimizer_state_dict"])

        self.current_epoch = checkpoint["epoch"] + 1
        self.best_validation_metric = checkpoint["best_loss"]

        if scheduler is not None:
            scheduler.load_state_dict(checkpoint["scheduler_state_dict"])

        if amp is not None:
            amp.load_state_dict(checkpoint["amp"])

        return model, optimizer, self.current_epoch, scheduler, amp



from simple_framework.backends.BaseBackendClass import BackendBase

from simple_framework.trainer.BaseTrainerClass import SimpleFrameworkWrapper
from typing import Dict

from tqdm import tqdm
import logging
import pandas as pd
import os

import torch

from simple_framework.utilities.metrics import AverageMeter
from simple_framework.utilities.checkpoint_saver import Checkpoint_saver


class SimpleBackend(BackendBase):
    def __init__(self, model: SimpleFrameworkWrapper):
        super().__init__(model)

    def setup(self, settings: Dict):
        self.optimizer, self.scheduler = self.model.get_optimizer_scheduler()
        self.current_epoch = 0
        self.metric_container = {}
        self.global_step = 0
        self.settings = settings

        self.checkpointer = Checkpoint_saver(
            checkpoints_dir=self.settings["save_path"],
            description=self.settings["description"],
        )

        self.set_logger()

        self.train_columns = ["epoch", "step", "current_loss"]
        self.validate_columns = ["epoch"]

        self.train_df = None
        self.valid_df = pd.DataFrame(columns=["epoch", "loss", "avg_metric"])

    def train_step(self, data, step):
        self.optimizer.zero_grad()
        loss, metrics = self.model.train_step(data)

        # creating metrics container first time we see it
        if not self.metric_container:
            super().initialize_csv_file(metrics.items())

        loss.backward()

        self.optimizer.step()

        if self.scheduler is not None:
            self.scheduler.step()

        for _, (key, val) in enumerate(metrics.items()):
            self.metric_container[key].update(val, self.settings["batch_size"])

        super().write_to_tensorboard(metrics.items())

        current_loss = self.metric_container["loss"].current
        avg_loss = self.metric_container["loss"].avg

        # Saving interval - optional
        if (
            self.settings["checkpoint_every_n_steps"] is not None
            and (step + 1) % self.settings["checkpoint_every_n_steps"] == 0
        ):
            self.checkpointer.save_checkpoint(
                metric_value=avg_loss,
                model=self.model,
                checkpoint_name=f"epoch_{self.current_epoch}_step_{step}_avg_loss_{avg_loss}.bin",
                checkpoint_path=os.path.join(self.settings["save_path"], f"epoch_{self.current_epoch}"),
            )

        # save best
        self.checkpointer.should_save_best(
            metric_value=self.metric_container["loss"].avg, model=self.model, task="train"
        )

        # adding given step data to csv file
        new_row = (
            [self.current_epoch, step, loss.detach().item()]
            + [metric.avg for metric in self.metric_container.values()]
            + [super().get_learning_rate()]
        )

        new_series = pd.Series(new_row, index=self.train_df.columns)

        self.train_df = self.train_df.append(new_series, ignore_index=True)
        self.train_df.to_csv(os.path.join(self.settings["experiment_path"], "train_logs.csv"), index=False)

        return current_loss

    def train_epoch(self, train_dataloader):

        self.model.train()
        torch.set_grad_enabled(True)

        epochs = self.settings["epochs"]

        dl_iter = iter(train_dataloader)

        pbar = tqdm(range(self.settings["steps_per_epoch"]), dynamic_ncols=True)

        for step in pbar:
            self.global_step = step + self.current_epoch * self.settings["steps_per_epoch"]
            try:
                data = next(dl_iter)
            except StopIteration:
                dl_iter = iter(train_dataloader)
                data = next(dl_iter)

            current_loss = self.train_step(data, step)

            avg_loss = self.metric_container["loss"].avg

            # set pbar description
            pbar.set_description(
                f"TRAIN epoch {self.current_epoch+1}/{epochs} idx {step} \
                current loss {current_loss}, avg loss {avg_loss}"
            )

        # Saving last checkpoint in epoch
        self.checkpointer.save(
            checkpoint_name="last_checkpoint.bin",
            model=self.model,
            optimizer=self.optimizer,
            scheduler=self.scheduler,
        )

    def train_phase(self, dataset: torch.utils.data.Dataset, validation_dataset: torch.utils.data.Dataset):

        train_dataloader = torch.utils.data.DataLoader(
            dataset,
            shuffle=self.settings["shuffle"],
            batch_size=self.settings["batch_size"],
            num_workers=self.settings["num_workers"],
        )

        if self.settings["steps_per_epoch"] is None:
            self.settings["steps_per_epoch"] = len(train_dataloader)

        for epoch in range(self.settings["epochs"]):
            logging.info("starting epoch {}/{} training step".format(epoch + 1, self.settings["epochs"]))
            self.checkpointer.set_epoch(self.current_epoch)
            self.train_epoch(train_dataloader)

            if validation_dataset is not None and (epoch + 1) % self.settings["validation_freq"] == 0:
                self.validation_phase(validation_dataset)

            self.current_epoch += 1

    def validation_phase(self, dataset: torch.utils.data.Dataset):

        validation_dataloader = torch.utils.data.DataLoader(
            dataset,
            shuffle=False,
            batch_size=self.settings["validation_batch_size"],
            num_workers=self.settings["num_workers"],
        )

        if self.settings["validation_steps"] is None:
            self.settings["validation_steps"] = len(validation_dataloader)

        epochs = self.settings["epochs"]
        validation_sum_loss = AverageMeter()
        val_metric = AverageMeter()
        self.model.eval()

        with torch.no_grad():
            dl_iter = iter(validation_dataloader)

            pbar = tqdm(range(self.settings["validation_steps"]), dynamic_ncols=True)

            for step in pbar:
                try:
                    data = next(dl_iter)
                except StopIteration:
                    dl_iter = iter(validation_dataloader)
                    data = next(dl_iter)

                loss, metrics = self.model.train_step(data)

                # add to average meter for loss
                validation_sum_loss.update(loss.item(), self.settings["validation_batch_size"])

                # validation metric update
                val_metric.update(metrics[self.settings["validation_metric"]], self.settings["validation_batch_size"])

                # set pbar description
                pbar.set_description(
                    (
                        f"VALIDATION epoch {self.current_epoch+1}/{epochs} step {step}"
                        f" current loss  {validation_sum_loss.current}, avg loss {validation_sum_loss.avg}"
                        f", validation_metric {val_metric.avg}"
                    )
                )

            # adding given epoch data to csv file
            new_row = [self.current_epoch, validation_sum_loss.avg, val_metric.avg]
            new_series = pd.Series(new_row, index=self.valid_df.columns)

            self.valid_df = self.valid_df.append(new_series, ignore_index=True)
            self.valid_df.to_csv(os.path.join(self.settings["experiment_path"], "validation_logs.csv"), index=False)
            logging.info(f"Validation result metric for epoch {self.current_epoch} = {val_metric.avg}")

            # save best
            self.checkpointer.should_save_best(
                metric_value=val_metric.avg,
                checkpoint_name="best_model_validation_{}.pth".format(self.settings["description"]),
                model=self.model,
                task="validation",
            )



from simple_framework.backends.simple_backend import SimpleBackend
from simple_framework.backends.horovod_backend import HorovodBackend

from simple_framework.utilities.checkpoint_saver import Checkpoint_saver
from simple_framework.utilities.metrics import AverageMeter

__all__ = ["SimpleBackend", "HorovodBackend", "Checkpoint_saver", "AverageMeter"]



__version__ = "0.0.1.7"

from simple_framework.trainer import Trainer, SimpleFrameworkWrapper

from simple_framework.utilities.schedulers import get_flat_cosine_schedule
from simple_framework.utilities.metrics import AverageMeter
from simple_framework.utilities.checkpoint_saver import Checkpoint_saver
from simple_framework.backends import SimpleBackend, HorovodBackend


__all__ = [
    "Trainer",
    "get_flat_cosine_schedule",
    "AverageMeter",
    "Checkpoint_saver",
    "SimpleBackend",
    "HorovodBackend",
]

# for compatibility with namespace packages
__import__("pkg_resources").declare_namespace(__name__)



import math
from torch.optim.lr_scheduler import LambdaLR

"""
    flat + cosine schedule
"""


def get_flat_cosine_schedule(optimizer, num_training_steps, percentege_of_const=0.7, num_cycles=0.5, last_epoch=-1):
    """
    Before percentage_of_const get constant lr,
    then do cosine decay till ends
    """
    # print("num_training_steps = ", num_training_steps)
    # print("int(num_training_steps*percentege_of_const) = ", int(num_training_steps*percentege_of_const))

    def lr_lambda(current_step):
        if current_step < int(num_training_steps * percentege_of_const):
            # print(current_step, " returning 1")
            return 1.0
        progress = float(current_step - num_training_steps * percentege_of_const) / float(
            max(1, num_training_steps - num_training_steps * percentege_of_const)
        )
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

    return LambdaLR(optimizer, lr_lambda, last_epoch)



from abc import ABC, abstractmethod
import torch


class SimpleFrameworkWrapper(ABC, torch.nn.Module):
    @abstractmethod
    def __init__(self, model: torch.nn.Module = None, loss_fn=None):
        super().__init__()
        assert loss_fn is not None
        assert model is not None
        self.model = model
        self.loss_fn = loss_fn

    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)

    @abstractmethod
    def train_step(self, batch_data):
        pass

    @abstractmethod
    def validation_step(self, batch_data):
        pass

    @abstractmethod
    def get_optimizer_scheduler(self):
        pass



from abc import ABC, abstractmethod
import torch
import os
import logging
import pandas as pd

from pathlib import Path

from simple_framework.trainer.BaseTrainerClass import SimpleFrameworkWrapper
from typing import Dict

from simple_framework.utilities.checkpoint_saver import Checkpoint_saver
from simple_framework.utilities.metrics import AverageMeter

from torch.utils.tensorboard import SummaryWriter

writer = SummaryWriter()


class BackendBase(ABC):
    def __init__(self, model: SimpleFrameworkWrapper):
        self.model = model

    @abstractmethod
    def setup(self, settings: Dict):
        raise NotImplementedError

    @abstractmethod
    def train_step(self, data, step):
        raise NotImplementedError

    @abstractmethod
    def train_epoch(self, data):
        raise NotImplementedError

    @abstractmethod
    def train_phase(self, dataset: torch.utils.data.Dataset):
        raise NotImplementedError

    @abstractmethod
    def validation_phase(self, dataset: torch.utils.data.Dataset):
        raise NotImplementedError

    def set_logger(self):
        logging_dir = os.path.join(self.settings["experiment_path"], "info.log")

        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)

        logger = logging.getLogger()

        fhandler = logging.FileHandler(filename=logging_dir, mode="a")
        formatter = logging.Formatter("%(asctime)s - %(process)d -  %(message)s")
        fhandler.setFormatter(formatter)
        fhandler.setLevel(logging.INFO)

        logger.addHandler(fhandler)
        logger.addHandler(ch)

    def initialize_csv_file(self, metrics_data):
        for i, (key, val) in enumerate(metrics_data):
            self.train_columns.append(key)
            self.metric_container[key] = AverageMeter()
            logging.info(f"adding metric {key}")
        self.train_columns.append("learning_rate")
        self.train_df = pd.DataFrame(columns=self.train_columns)

    def get_learning_rate(self):
        lr = []
        for param_group in self.optimizer.param_groups:
            lr += [param_group["lr"]]
        return lr

    def write_to_tensorboard(self, metrics_data):
        # filling metrics/loss and adding to tensorboard

        description = self.settings["description"]

        for _, (key, val) in enumerate(metrics_data):
            if key == "loss":
                continue
            writer.add_scalar(f"Metric/{key}_{description}", self.metric_container[key].avg, self.global_step)
            writer.add_scalar(f"Metric/{key}_all_runs", self.metric_container[key].avg, self.global_step)

        # writer to tensorboard
        writer.add_scalar(f"Loss/train_{description}", self.metric_container["loss"].avg, self.global_step)

        writer.add_scalar("Loss/all_train_losses", self.metric_container["loss"].avg, self.global_step)

        # adding all learning rates to tensorboard
        for idx, lr_value in enumerate(self.get_learning_rate()):
            writer.add_scalar(f"Learning_rates/lr_{idx}_{description}", lr_value, self.global_step)



class AverageMeter:
    """Computes and stores the average and current value"""

    def __init__(self, batch_accumulation=1):
        self.batch_accumulation = batch_accumulation
        self.reset()

    def reset(self):
        self.current = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.current = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / (self.count * self.batch_accumulation)
